/*
  mergemod_20rha.c
  Kernel module for Red Hat Alpha linux
  
  This file is part of mergemem by Philipp Richter & Philipp Reisner

  mergemem is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  mergemem is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with mergemem; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/swap.h>

#include <asm/page.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/pgtable.h>

#include "mergemod.h"

#define MODULE_NAME "mergemod :"
#define MERGEMEM_INODE 3711

static long mergemod_read(struct inode*, struct file*, char*,unsigned long);
static long mergemod_write(struct inode*, struct file*, const char*,
			   unsigned long);
static int mergemod_open(struct inode*,struct file*);
static void mergemod_close(struct inode*,struct file*);
static int mergemod(int, unsigned long, int, unsigned long );
static int genchksum(int, unsigned long, unsigned long*, int* );

static struct file_operations proc_mergemod_file_operations = 
{
  NULL,                   /* lseek - default */
  mergemod_read,             /* read */
  mergemod_write,            /* write  */
  NULL,                   /* readdir */
  NULL,                   /* select - default */
  NULL,                   /* ioctl - default */
  NULL,                   /* mmap */
  mergemod_open,          /* open code */
  mergemod_close,         /* release code */
  NULL                    /* can't fsync */
};

static struct inode_operations proc_mergemod_inode_operations =
{
  &proc_mergemod_file_operations,   /* my own file-ops */
  NULL,                   /* create */
  NULL,                   /* lookup */
  NULL,                   /* link */
  NULL,                   /* unlink */
  NULL,                   /* symlink */
  NULL,                   /* mkdir */
  NULL,                   /* rmdir */
  NULL,                   /* mknod */
  NULL,                   /* rename */
  NULL,                   /* readlink */
  NULL,                   /* follow_link */
  NULL,                   /* readpage */
  NULL,                   /* writepage */
  NULL,                   /* bmap */
  NULL,                   /* truncate */
  NULL                    /* permission */
};
  

struct proc_dir_entry proc_mergemod = 
{
  MERGEMEM_INODE, 8, "mergemem",
  S_IFREG | S_IRUGO | S_IWUGO, 1, 0, 0,
  0, &proc_mergemod_inode_operations,
  NULL, NULL,
  NULL,
  NULL, NULL
};


int init_module()
{
  printk(MODULE_NAME"init_module()\n");

  proc_register(&proc_root,&proc_mergemod);
  
  return 0;
}

void cleanup_module()
{
  printk(MODULE_NAME"cleanup_module()\n");
  proc_unregister(&proc_root,MERGEMEM_INODE);
}

/****************************************************************************/
/* From here on, we have the interface to the user level process            */

char mergemod_test_message[] = "This is the kernel entry point of the mergemem"
                            "project\nthere is nothing of interest here.\n";

static long mergemod_read(struct inode* node, struct file* file, char* buf, unsigned long count)
{
  int fpos = file->f_pos;

  if(fpos > sizeof(mergemod_test_message)-1) return 0;
  if((sizeof(mergemod_test_message)-fpos) < count) 
    count=sizeof(mergemod_test_message)-fpos;
  memcpy_tofs(buf,mergemod_test_message+fpos,count);
  file->f_pos = file->f_pos+count;
  return count;
}

static long mergemod_write(struct inode* node, struct file* file, const char* buf, unsigned long count)
{
  struct mergemod_struct mms;
  int r;

  if(count != sizeof(struct mergemod_struct))
    {
      printk(MODULE_NAME"request of illegal size\n");
      return MERGEMEM_ILLREQ;
    }

  memcpy_fromfs(&mms, buf, count);

  if(mms.pid2 == 0) 
    {
      r = genchksum(mms.pid1,mms.addr1,&mms.addr2,&mms.pid2);
      memcpy_tofs((char *)buf, &mms, count);
      return r;
    }
  else  
    return mergemod(mms.pid1,mms.addr1,mms.pid2,mms.addr2);
}

static int mergemod_open(struct inode* i,struct file* f)
{
  MOD_INC_USE_COUNT;
  return 0;
}

static void mergemod_close(struct inode* i,struct file* f)
{
  MOD_DEC_USE_COUNT;
}

/****************************************************************************/
/* From here on, we have the implementation of the page fusion function     */


static struct task_struct * get_task(int pid)
{
  struct task_struct * tsk, * st;

  st = tsk = current_set[0];
  
  do
    {
      if(tsk->pid == pid) return tsk;
      tsk=tsk->next_task;
    }
  while(tsk != st);
  
  return 0;
}

static char * get_phys_addr(struct task_struct * tsk,unsigned long addr,
			    pte_t ** pe)
{
  pgd_t * page_dir;
  pmd_t * page_middle;
  pte_t pte;

  page_dir = pgd_offset(tsk->mm,addr);
  if (pgd_none(*page_dir)) return (char *)0;
  if (pgd_bad(*page_dir)) 
    {
      printk("Bad page dir entry %08lx\n", pgd_val(*page_dir));
      pgd_clear(page_dir);
      return (char *)0;
    }
  
  page_middle = pmd_offset(page_dir,addr);
  if (pmd_none(*page_middle)) return (char *)0;
  if (pmd_bad(*page_middle)) 
    {
      printk("Bad page middle entry %08lx\n", pmd_val(*page_middle));
      pmd_clear(page_middle);
      return (char *)0;
    }
  
  pte = *pte_offset(page_middle,addr);
  if (!pte_present(pte)) return (char *)0;

  *pe = pte_offset(page_middle,addr);

  return (char *) pte_page(pte);
}

static int mergemod(int pid1,unsigned long addr1,int pid2, unsigned long addr2)
{
  struct task_struct * tsk1, * tsk2;
  char * page1, * page2;
  pte_t * pte1, * pte2, pte;
  
  tsk1 = get_task(pid1);
  if(!tsk1) return MERGEMEM_NOTASK1;
  tsk2 = get_task(pid2);
  if(!tsk2) return MERGEMEM_NOTASK2;
  
  page1 = get_phys_addr(tsk1,addr1,&pte1);
  if(!page1) return MERGEMEM_NOPAGE1;
  page2 = get_phys_addr(tsk2,addr2,&pte2);
  if(!page2) return MERGEMEM_NOPAGE2;
  
  if(page1 == page2) 
    return MERGEMEM_ALREADYSH;

  if(mem_map[MAP_NR(page2)].count != 1) 
     return MERGEMEM_MOREMAP;

  cli(); /* start of critical section */
  if(memcmp(page1,page2,PAGE_SIZE))
    {
      sti(); /* possible end of critical section */
      return MERGEMEM_NOTEQUAL;
    }
  
  /* Do the actual work... */

  /*  new pte's are readonly and point to page1 */
  pte = pte_wrprotect(*pte1);

  /*  increase the count in the according mem_map structure */
  mem_map[MAP_NR(page1)].count++;

  /* page1 must go off the swap_cache, since it is now mapped more
   * than one time
   */
  if(delete_from_swap_cache(MAP_NR(page1)))
    pte = pte_mkdirty(pte);

  set_pte(pte1,pte);
  set_pte(pte2,pte_mkold(pte));

  sti(); /* possible end of critical section */  
  /*  free page2 */
  free_page((unsigned long)page2);

  /*  decrease the rss counters */
  tsk2->mm->rss--;

  /* No, TLB flushing needed, because the processes which PTEs where
   * altered, are not running at the moment.
   */
  return MERGEMEM_SUCCESS;
}

/* Calculate checksum for page */
static int genchksum(int pid, unsigned long addr, unsigned long* ret_chksum,
		     int* ret_mappings)
{
  struct task_struct * tsk;
  unsigned long * page, * end;
  pte_t * pe;
  unsigned long chksum;
    
  tsk = get_task(pid);
  if(!tsk) return MERGEMEM_NOTASK1;

  page = (unsigned long *) get_phys_addr(tsk,addr,&pe);
  if(!page) return MERGEMEM_NOPAGE1;

  chksum=0;
  /* since this are unsigned long* pointers we must devide PAGE_SIZE
   * by the size of unsigned long (PAGE_SIZE is the size of a page in bytes)
   */
  /* For enhanced performance, we use only the first half of a page
   * for the checksum 
   */  
  *ret_mappings = mem_map[MAP_NR(page)].count;

  end = page + (PAGE_SIZE/sizeof(unsigned long))/2;     
  while(page < end)
    {
      chksum = chksum ^ *page;
      page++;
    }

  *ret_chksum = chksum;

  return MERGEMEM_SUCCESS;
}

